@@ -1,8 +1,10 @@
 from ._version import __version__
 
 from .argcheckers import Any, Is, Option, Seq, AllOf, Dict, Union, Seq, Eq, \
-    Geq, Gt, IsIterable, Leq, Lt, Neq, Set
+    Geq, Gt, Leq, Lt, Neq, Set
 
 from .contracts import check_sig, contracts, precondition, postcondition, \
     invariant
 
+from .types import NamedTuple
+
