@@ -1,5 +1,6 @@
 import inspect
 from collections import abc
+from .error import CheckError
 
 from codeviking.contracts.error import PreconditionError, \
     PreconditionViolation, InvariantError, SignatureError, SignatureViolation, \
@@ -12,24 +13,20 @@ from .error import AnnotationError, CheckViolation, \
 
 
 class Checker:
-    pass
-
-
-class CheckerError(Exception):
-    def __init__(self, *,
-                 message,
-                 arg):
-        self.message = message
-        self.arg = arg
-        super().__init__(message)
+    def __init__(self, to_str_func):
+        self._to_str_func = to_str_func
 
     def __str__(self):
-        return self.message
+        return self._to_str_func(self)
+
+    def __call__(self, arg, eval_scope):
+        pass
 
 
-class ArgChecker(Checker):
+class UnscopedChecker(Checker):
     """
-    Base class for all contract types.
+    Checker that uses a function to validate an argument.  The function takes
+    only the argument, not an evaluation scope.
     """
 
     def __init__(self, func, to_str_func=None):
@@ -41,27 +38,39 @@ class ArgChecker(Checker):
 
         :param func: argument checking function that must be satisfied.
         :type func:  f(arg) -> bool
+        :param to_str_func: function that returns a string representation of
+            this object, or None if func.__name__ should be used.
+        :type to_str_func: None | (UnscopedChecker) -> str
         """
+        if to_str_func is None:
+            if hasattr(func, '__name__'):
+                to_str_func = lambda s: s._func.__name__
+            else:
+                to_str_func = lambda s: str(s._func)
+        super().__init__(to_str_func)
         self._func = func
-        self._to_str_func = to_str_func
+        if not callable(self._func):
+            raise ValueError(
+                "Cannot create an UnscopedChecker from func={func}.  "
+                "func must be a function that takes one "
+                "argument.".format(func=func))
 
-    def __call__(self, arg):
+    def __call__(self, arg, eval_scope):
         try:
             result = self._func(arg)
         except Exception as e:
             if not isinstance(e, TypeCheckViolation):
-                raise CheckerError(message="An exception was raised while "
-                                           "evaluating ArgChecker function: "
-                                           "{func}\n"
-                                           "on argument: {arg}\n"
-                                           "If this is a Checker in the "
-                                           "contracts package, please file a "
-                                           "bug "
-
-                                           "report.\n".format(
+                raise CheckError(message="An exception was raised while "
+                                         "evaluating UnscopedChecker function: "
+                                         "{func}\n"
+                                         "on argument: {arg}\n"
+                                         "If this is a Checker in the "
+                                         "contracts package, please file a "
+                                         "bug "
+                                         "report.\n".format(
                     func=self._func.__name__,
                     arg=str(arg)),
-                                   arg=arg) from e
+                    arg=arg) from e
             else:
                 raise
         if result:
@@ -69,16 +78,68 @@ class ArgChecker(Checker):
         # The function returned False.  We'll raise a CheckViolation for it.
         raise FunctionCheckViolation(function=self._func, value=arg)
 
-    def __str__(self):
-        if self._to_str_func is None:
-            return str(self._func.__name__)
-        return self._to_str_func(self)
+
+class ScopedChecker(Checker):
+    """
+    Checker that uses a function to validate an argument.  The function takes
+    the argument to validate and a scope in which to evaluate it.
+    """
+
+    def __init__(self, func, to_str_func=None):
+        """
+        Uses the function `func` to validate an argument.
+
+        When an arg or return value needs to be validated,
+        func(arg, eval_scope) is called.  If the function returns True,
+        the condition is satisfied.
+
+        :param func: argument checking function that must be satisfied.
+        :type func:  f(arg, eval_scope) -> bool
+        :param to_str_func: function that returns a string representation of
+            this object, or None if func.__name__ should be used.
+        :type to_str_func: None | (UnscopedChecker) -> str
+        """
+        if to_str_func is None:
+            if hasattr(func, '__name__'):
+                to_str_func = lambda s: s._func.__name__
+            else:
+                to_str_func = lambda s: str(s._func)
+        super().__init__(to_str_func)
+        self._func = func
+        if not callable(self._func):
+            raise ValueError(
+                "Cannot create an ScopedChecker from func={func}.  "
+                "func must be a function that takes two "
+                "arguments.".format(func=func))
+
+    def __call__(self, arg, eval_scope):
+        try:
+            result = self._func(arg, eval_scope)
+        except Exception as e:
+            if not isinstance(e, CheckViolation):
+                raise CheckError(message="An exception was raised while "
+                                         "evaluating ScopedChecker function: "
+                                         "{func}\n"
+                                         "on argument: {arg}\n"
+                                         "If this is a Checker in the "
+                                         "contracts package, please file a "
+                                         "bug "
+                                         "report.\n".format(
+                    func=self._func.__name__,
+                    arg=str(arg)),
+                    arg=arg) from e
+            else:
+                raise
+        if result:
+            return result
+        # The function returned False.  We'll raise a CheckViolation for it.
+        raise FunctionCheckViolation(function=self._func, value=arg)
 
 
-Any = ArgChecker(lambda x: True, lambda slf: 'Any')
+Any = UnscopedChecker(lambda x: True, lambda slf: 'Any')
 
 
-class Is(ArgChecker):
+class Is(UnscopedChecker):
     """
     ArgChecker that is satisfied when an arg is the same object as another.
     """
@@ -99,7 +160,7 @@ class Is(ArgChecker):
         return "Is(%s)" % str(self._compare_to)
 
 
-class Option(ArgChecker):
+class Option(ScopedChecker):
     """
     An ArgChecker that is satisfied when the argument satisfies its
     inner_checker, or when the argument is None
@@ -108,19 +169,16 @@ class Option(ArgChecker):
     def __init__(self, inner_checker):
         inner_checker = make_arg_checker(inner_checker)
         self._inner_checker = inner_checker
-        super().__init__(lambda x: x is None or inner_checker(x))
+        super().__init__(lambda x, eval_scope: x is None or
+                                               inner_checker(x, eval_scope))
 
     def __str__(self):
         return "Option(%s)" % str(self._inner_checker)
 
 
-class IsIterable(ArgChecker):
-    pass
-
-
 # Start Comparison Checkers
 
-class Cmp(ArgChecker):
+class Cmp(UnscopedChecker):
     def __init__(self, value, cmp_func, format_string):
         super().__init__(lambda x: cmp_func(x, value))
         self._value = value
@@ -166,20 +224,20 @@ class Geq(Cmp):
 
 
 
-class Iterable(ArgChecker):
+class Iterable(ScopedChecker):
     def __init__(self, element_checker, expected_type=abc.Iterable):
         """
-        An ArgChecker that is satisfied if the arg is xterable,
-        and all the elements of the list satisfy `element_checker`
+        A Checker that is satisfied if the arg is iterable, and all the
+        elements of the list satisfy `element_checker`
         """
         self._element_checker = make_arg_checker(element_checker)
 
-        def f(x):
+        def f(x, eval_scope):
             if not isinstance(x, expected_type):
                 raise TypeCheckViolation(value=x,
                                          expected_type=expected_type)
             for e in x:
-                if not self._element_checker(e):
+                if not self._element_checker(e, eval_scope):
                     raise ElementCheckViolation(self._element_checker, e)
             return True
 
@@ -203,7 +261,7 @@ class Set(Iterable):
         return "Set(%s)" % (str(self._element_checker))
 
 
-class Dict(ArgChecker):
+class Dict(ScopedChecker):
     def __init__(self, key_checker, value_checker):
         """
         An ArgChecker that is satisfied if the arg is an instance of dict,
@@ -213,13 +271,13 @@ class Dict(ArgChecker):
         self._key_checker = make_arg_checker(key_checker)
         self._value_checker = make_arg_checker(value_checker)
 
-        def f(d):
+        def f(d, eval_scope):
             if not isinstance(d, dict):
                 raise TypeCheckViolation(value=d, expected_type=dict)
             for (k, v) in d.items():
-                if not self._key_checker(k):
+                if not self._key_checker(k, eval_scope):
                     return False
-                if not self._value_checker(v):
+                if not self._value_checker(v, eval_scope):
                     return False
             return True
 
@@ -230,7 +288,7 @@ class Dict(ArgChecker):
                              str(self._value_checker))
 
 
-class Union(ArgChecker):
+class Union(ScopedChecker):
     def __init__(self, *arg_checkers):
         """
         An ArgChecker that is satisfied if one or more of the ArgCheckers in
@@ -238,13 +296,13 @@ class Union(ArgChecker):
         """
         self._all_checkers = [make_arg_checker(c) for c in arg_checkers]
 
-        def f(x):
+        def f(x, eval_scope):
             if len(self._all_checkers) == 0:
                 return True
             violations = list()
             for c in self._all_checkers:
                 try:
-                    if c(x):
+                    if c(x, eval_scope):
                         return True
                     else:
                         violations.append((c, False))
@@ -263,7 +321,6 @@ class Union(ArgChecker):
             raise UnionCheckViolation(message_lines=message,
                                       value=x)
 
-
         super().__init__(f)
 
     def __str__(self):
@@ -277,7 +334,7 @@ class Union(ArgChecker):
 # `make_arg_checker` from the appropriate annotations.
 
 
-class AllOf(ArgChecker):
+class AllOf(ScopedChecker):
     def __init__(self, *arg_checkers):
         """
         An ArgChecker that is satisfied if all of the ArgCheckers in
@@ -285,9 +342,9 @@ class AllOf(ArgChecker):
         """
         self._all_checkers = [make_arg_checker(c) for c in arg_checkers]
 
-        def f(x):
+        def f(x, eval_scope):
             for c in self._all_checkers:
-                if not c(x):
+                if not c(x, eval_scope):
                     return False
             return True
 
@@ -297,7 +354,7 @@ class AllOf(ArgChecker):
         return "AllOf(%s)" % (", ".join([str(c) for c in self._all_checkers]))
 
 
-class TypeArgChecker(ArgChecker):
+class TypeArgChecker(ScopedChecker):
     def __init__(self, ctype):
         """
         An ArgChecker that is satisfied if the argument is an instance of
@@ -305,7 +362,7 @@ class TypeArgChecker(ArgChecker):
         """
         self._ctype = ctype
 
-        def check_type(value):
+        def check_type(value, eval_scope):
             if isinstance(value, ctype):
                 return True
             raise TypeCheckViolation(value=value, expected_type=ctype)
@@ -316,35 +373,105 @@ class TypeArgChecker(ArgChecker):
         return self._ctype.__name__
 
 
-class TupleArgChecker(ArgChecker):
+class TupleArgChecker(ScopedChecker):
     def __init__(self, checkers):
         """
         An ArgChecker that is satisfied if the argument is an ordered sequence
-        whose ith element is an instance of the ith types in `ctypes`
+        whose ith element satisfies the ith checker.
         """
-        super().__init__(None)
         arg_checkers = tuple(make_arg_checker(a) for a in checkers)
         self._arg_checkers = arg_checkers
 
-    def __call__(self, arg):
-        expected_length = len(self._arg_checkers)
-        try:
-            arg_len = len(arg)
-        except TypeError as e:
-            raise TypeCheckViolation(value=arg, expected_type=Seq)
-        if expected_length != arg_len:
-            raise LengthMismatchViolation(value=arg,
-                                          expected_length=expected_length)
-        try:
-            for i in range(len(self._arg_checkers)):
-                if not self._arg_checkers[i](arg[i]):
-                    return False
-            return True
-        except CheckViolation as e:
-            raise
+        def f(arg, eval_scope):
+            expected_length = len(self._arg_checkers)
 
-    def __str__(self):
-        return "(%s)" % (", ".join([str(c) for c in self._arg_checkers]))
+            try:
+                arg_len = len(arg)
+            except TypeError as e:
+                raise TypeCheckViolation(value=arg, expected_type=Seq)
+            if expected_length != arg_len:
+                raise LengthMismatchViolation(value=arg,
+                                              expected_length=expected_length)
+            try:
+                for i in range(len(self._arg_checkers)):
+                    r = self._arg_checkers[i](arg[i], eval_scope)
+                    if not r:
+                        return False
+            except CheckViolation as e:
+                raise
+            return True
+
+        def to_str(s):
+            return "(%s)" % (", ".join([str(c) for c in s._arg_checkers]))
+
+        super().__init__(f, to_str)
+
+
+class NamedTupleArgChecker(ScopedChecker):
+    def __init__(self, elements):
+        """
+        An ArgChecker that is satisfied if the argument
+        whose ith element satisfies the ith checker.
+        """
+        self._arg_names = tuple(n for n, c in elements)
+        self._arg_checkers = tuple(make_arg_checker(c) for n, c in elements)
+        self._kwarg_checkers = {n: make_arg_checker(c) for n, c in elements}
+
+        def f(arg, eval_scope):
+            expected_length = len(self._arg_checkers)
+            args, kwargs = arg
+            try:
+                arg_len = len(args) + len(kwargs)
+            except TypeError as e:
+                raise TypeCheckViolation(value=(args, kwargs),
+                                         expected_type=Seq)
+            if expected_length != arg_len:
+                raise LengthMismatchViolation(value=(args, kwargs),
+                                              expected_length=expected_length)
+            try:
+                for a, checker in zip(args, self._arg_checkers):
+                    r = checker(a, eval_scope)
+                    if not r:
+                        return False
+                for name in kwargs:
+                    r = self._kwarg_checkers[name](kwargs[name], eval_scope)
+                    if not r:
+                        return False
+            except CheckViolation as e:
+                raise
+            return True
+
+        def to_str(s):
+            return "(%s)" % (", ".join(["%s: %s" % (n, str(c)) for n, c in
+                                        zip(s._arg_names, s._arg_checkers)]))
+
+        super().__init__(f, to_str)
+
+
+class DelayedEvalChecker(ScopedChecker):
+    """
+    A Checker that is not built until the first time it checks an argument.
+    """
+
+    def __init__(self, expression):
+        if not isinstance(expression, str):
+            raise ValueError("expression must be a string.")
+        self._expression = expression
+        self._checker = None
+
+        def f(arg, eval_scope):
+            if self._checker is None:
+                # checker has not been evaluated yet, so we do that now.
+                try:
+                    self._checker = make_arg_checker(eval(self._expression,
+                                                          eval_scope))
+                except Exception as e:
+                    raise CheckError("DelayedEvalChecker is unable to evaluate "
+                                     "expression %s " % str(
+                        self._expression)) from e
+            return self._checker(arg, eval_scope)
+
+        super().__init__(f, lambda s: s._expression)
 
 
 # End internal arg checkers
@@ -359,12 +486,16 @@ def make_arg_checker(annotation):
 
     :param annotation: the argument annotation
     :type annotation: ArgChecker, None, type, callable, str, set, tuple, list
+    :param eval_scope: the scope in which to evaluate the annotation.  This is
+        only relevant for checkers that must be lazily evaluated (like
+        "ClassName").  Usually, this will be the scope in which the function
+        is defined.
     :return: an ArgChecker
     :rtype: function that takes any argument and returns a boolean if the
             argument satisfies the checker.
 
     """
-    if isinstance(annotation, ArgChecker):
+    if isinstance(annotation, Checker):
         return annotation
     if annotation is None:
         return Is(None)
@@ -382,11 +513,11 @@ def make_arg_checker(annotation):
     if isinstance(annotation, list):
         return AllOf(*annotation)
     if isinstance(annotation, str):
-        return make_arg_checker(eval(annotation))
+        return DelayedEvalChecker(annotation)
     if isinstance(annotation, set):
         return Union(*annotation)
     if callable(annotation):
-        return ArgChecker(annotation)
+        return UnscopedChecker(annotation)
     raise ContractError("Invalid contract '%s'" % str(annotation))
 
 
@@ -435,6 +566,7 @@ class ContractWrapper:
             self._return_check = make_arg_checker(
                 self._signature.return_annotation)
 
+        self._eval_scope = self._target.__globals__
         # build a checker for each non-empty parameter annotation
         for name, p in self._signature.parameters.items():
             if p.annotation is not inspect.Signature.empty:
@@ -442,7 +574,6 @@ class ContractWrapper:
         if len(self._arg_checkers) == 0:
             self._arg_checkers = None
 
-
     def __call__(self, *args, **kwargs):
 
         for prc in self._preconditions:
@@ -480,7 +611,7 @@ class ContractWrapper:
                     continue
                 checker = self._arg_checkers[arg_name]
                 try:
-                    success = checker(arg_value)
+                    success = checker(arg_value, self._eval_scope)
                 except TypeCheckViolation:
                     raise
                 except Exception as e:
@@ -499,11 +630,14 @@ class ContractWrapper:
         # otherwise, it will cause the check_sig on this function to fail).
         return_value = self._target(*args, **kwargs)
 
+
         # now we can check the return value
         if self._return_check is not None:
             try:
                 # noinspection PyCallingNonCallable
-                return_success = self._return_check(return_value)  # TODO
+                return_success = self._return_check(return_value,
+                                                    self._eval_scope)
+                # TODO
             except Exception as e:
                 raise SignatureError(target=self._target,
                                      arg_name='return',
@@ -539,7 +673,7 @@ class ContractWrapper:
                                          pre_val=pre_invariants[i],
                                          post_val=post_invariants[i],
                                          return_value=return_value
-                )  # TODO
+                                         )  # TODO
         for postc in self._postconditions:
             try:
                 r = postc(*args, **kwargs)
