@@ -1,7 +1,6 @@
-from collections import namedtuple
-
-from codeviking.contracts.argcheckers import TupleArgChecker
+from collections import namedtuple, OrderedDict
 
+from codeviking.contracts.argcheckers import NamedTupleArgChecker
 
 __author__ = 'dan'
 
@@ -16,16 +15,13 @@ def NamedTuple(name, elements):
     :return: the class
     :rtype: namedtuple
     """
-    # TODO fix this so keyword parameters can be used in the constructor
-    elem_names = [n for (n, c) in elements]
-    checkers = [c for (n, c) in elements]
-    checker = TupleArgChecker(checkers)
-    T = namedtuple(name, elem_names)
+    checker = NamedTupleArgChecker(elements)
+    T = namedtuple(name, [n for (n, _) in elements])
     orig_new = T.__new__
 
-    def __new__(*args):
-        ok = checker(args[1:])
-        return orig_new(*args)
+    def __new__(*args, **kwargs):
+        ok = checker((args[1:], kwargs), T.__new__.__globals__)
+        return orig_new(*args, **kwargs)
 
     T.__new__ = __new__
     T.__name__ = name
