@@ -30,7 +30,7 @@ def test_arg_str(arg_str):
 def test_arg_success(arg_success):
     arg, success = arg_success
     a = make_arg_checker(arg)
-    assert a(success)
+    assert a(success, globals())
 
 
 def test_arg_violation(arg_violation):
@@ -38,7 +38,7 @@ def test_arg_violation(arg_violation):
     assert exception is not None
     a = make_arg_checker(arg)
     try:
-        r = a(violation)
+        _ = a(violation, globals())
     except Exception as e:
         if isinstance(e, exception):
             assert True
